# Generated by Django 4.0.6 on 2022-08-02 18:31

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('urlid_graph', '0004_alter_joblog_id_alter_logstep_message'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='joblog',
            options={'ordering': ('-created_at',)},
        ),
        migrations.AlterModelOptions(
            name='logstep',
            options={'ordering': ('-created_at',)},
        ),
    ]
