# Generated by Django 4.0.6 on 2022-08-02 09:44

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('urlid_graph', '0003_joblog_jobstep'),
    ]

    operations = [
        migrations.AlterField(
            model_name='joblog',
            name='id',
            field=models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='logstep',
            name='message',
            field=models.TextField(blank=True, null=True),
        ),
    ]
